"""Insertion Device Feedforward Configuration."""

import numpy as _np

from ..search import IDSearch as _IDSearch
from ..clientconfigdb import ConfigDBDocument as _ConfigDBDocument


class IDFFConfig:
    """."""

    CONFIGDB_TYPE = 'si_idff'

    def __init__(self, name=None):
        """."""
        self._name = None
        self._configdbdoc = None
        self._value = None
        if name:
            self.load_config(name=name)

    @property
    def name(self):
        """Return configuration name."""
        return self._name

    @name.setter
    def name(self, value):
        """Set configuration name."""
        self._name = value

    @property
    def kparameter_pvname(self):
        """Return ID kparameter pvname."""
        config = self._value
        if config:
            kparm = config['pvnames']['kparameter']
            return kparm
        else:
            raise ValueError('Configuration not loaded!')

    @property
    def ch_pvnames(self):
        """Return CH corrector power supply pvnames."""
        return self._get_corr_pvnames('ch1', 'ch2')

    @property
    def cv_pvnames(self):
        """Return CV corrector power supply pvnames."""
        return self._get_corr_pvnames('cv1', 'cv2')

    @property
    def qs_pvnames(self):
        """Return QS corrector power supply pvnames."""
        return self._get_corr_pvnames('qs1', 'qs2')

    @property
    def polarizations(self):
        """Return list of light polarizations in the IDFF config."""
        if self._value:
            return list(self._value['polarizations'].keys())
        else:
            raise ValueError('Configuration not loaded!')

    @property
    def value(self):
        """Return ID feedforward table."""
        return self._value

    @value.setter
    def value(self, value):
        """Set ID feedforward table."""
        # TODO: check dictionary
        self._value = value

    def calculate_setpoints(self, polarization, kparameter):
        """Return correctors setpoints for a particular ID config.

        The parameter 'kparameter' can be a gap or phase value,
        depending on the insertion device.
        """
        if self._value:
            setpoints = dict()
            idff = self._value['polarizations'][polarization]
            kparameters = idff['kparameter']
            setpoints = dict()
            for key, value in idff.items():
                if key != 'kparameter':
                    # linear interpolation
                    setpoint = _np.interp(kparameter, kparameters, value)
                    corr_pvname = self._value['pvnames'][key]
                    setpoints[corr_pvname] = setpoint
            return setpoints
        else:
            raise ValueError('Configuration not loaded!')

    @staticmethod
    def create_template_config(idname):
        """Create a template configuration for a given ID type."""
        idff = _IDSearch.conv_idname_2_idff(idname)

        # build value for pvnames key
        pvnames = dict()
        for key, corr_pvname in idff.items():
            if key != 'polarizations':
                pvnames[key] = corr_pvname

        # build value for polarizations key
        table = dict(kparameter=[0, ])
        for pvname in pvnames:
            table[pvname] = [0, ]
        polarizations = dict()
        for polarization in idff['polarizations']:
            polarizations[polarization] = dict(table)

        template_config = dict(pvnames=pvnames, polarizations=polarizations)
        return template_config

    def load_config(self, name=None):
        """Load IDFF configuration."""
        if name is not None:
            self._name = name
        self._configdbdoc = _ConfigDBDocument(
            config_type=IDFFConfig.CONFIGDB_TYPE, name=self._name)
        self._configdbdoc.load()
        self._value = self._configdbdoc.value

    def save_config(self, name=None):
        """Save IDFF configuration."""
        if name is not None:
            self._name = name
        self._configdbdoc = _ConfigDBDocument(
            config_type=IDFFConfig.CONFIGDB_TYPE, name=self._name)
        self._configdbdoc.value = self._value
        self._configdbdoc.save()

    def _get_corr_pvnames(self, cname1, cname2):
        """Return corrector power supply pvnames."""
        if self._value:
            pvnames = self._value['pvnames']
            corr1, corr2 = pvnames.get(cname1), pvnames.get(cname2)
            return corr1, corr2
        else:
            raise ValueError('Configuration not loaded!')



"""SI Insertion Device Feedforward (IDFF) Configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# IDFF corrector table
#
# 1st-level keys:
#   - 'pvnames': its value is a dict with
#       ('kparameter', 'ch1', 'ch2', ... 'qs1', 'qs2') as
#       keys and corresponding pvname values for
#       ID k-variation parameter readout and correctors setpoints.
#   - 'polarizations': its value is a dict with
#       ('horizontal', 'vertical', ...) light polarization configs of the ID
#       as keys and dicts with kparameter and correctors
#       IDFF tables as lists.

_template_dict = {
    'pvnames': {
        'kparameter': 'SI-10SB:ID-EPU50:Gap-Mon',
        'ch1': 'SI-10SB:PS-CH-1:Current-SP',
        'ch2': 'SI-10SB:PS-CH-2:Current-SP',
        'cv1': 'SI-10SB:PS-CV-1:Current-SP',
        'cv2': 'SI-10SB:PS-CV-2:Current-SP',
        'qs1': 'SI-10SB:PS-QS-1:Current-SP',
        'qs2': 'SI-10SB:PS-QS-2:Current-SP',
    },
    'polarizations': {
        'horizontal': {
            'kparameter': [-1, 0, 1],
            'ch1': 3*[0],
            'ch2': 3*[0],
            'cv1': 3*[0],
            'cv2': 3*[0],
            'qs1': 3*[0],
            'qs2': 3*[0],
            },
        'vertical': {
            'kparameter': [-2, -1, 0, 1, 2],
            'ch1': 5*[0],
            'ch2': 5*[0],
            'cv1': 5*[0],
            'cv2': 5*[0],
            'qs1': 5*[0],
            'qs2': 5*[0],
            },
    },
}
